/*
  A and B are inputs, O is out
  OP is the operation to perform:
    000 = add
    001 = sub
    010 = mul_lo
    011 = mul_hi
    (011)
    100 = and
    101 = or
    110 = xor
    111 = not A
*/
module alu(
  input CLK,
  input EN,
  output reg RDY,

  input [31:0] A,
  input [31:0] B,
  input [2:0] op,
  output wire [31:0] O_lo,
  output wire [31:0] O_hi, // only used for OP_MUL
  output wire Fflow,
  output wire Fzero
);

// Constants
/* verilator lint_off UNUSEDPARAM */
localparam OP_ADD = 3'b000;
localparam OP_SUB = 3'b001;
localparam OP_MUL = 3'b010;
localparam OP_AND = 3'b100;
localparam OP_OR  = 3'b101;
localparam OP_XOR = 3'b110;
localparam OP_NOT = 3'b111;

localparam ST_IDLE = 0;
localparam ST_MULTIPLY_START = 1;
localparam ST_MULTIPLY_END = 32;

localparam OUT_ADD = 0;
localparam OUT_SUB = 1;
localparam OUT_BW  = 2;
localparam OUT_MUL = 3;

// State

reg [8:0] state = ST_IDLE;
reg [2:0] selected_out;

assign RDY = state == ST_IDLE;

reg [31:0] bitwise_out;

// Multiplication

reg [63:0] mult_out;
reg [31:0] factorA, factorB; // factorA is static, factorB is shifted each cycle

// Outputs

assign O_lo =
  (selected_out == OUT_ADD || selected_out == OUT_SUB) ? adder_out :
  selected_out == OUT_MUL ? mult_out[31:0] :
  bitwise_out;

assign O_hi =
  selected_out == OUT_MUL ? mult_out[63:32] :
  0;

assign Fflow =
  selected_out == OUT_ADD ? adder_carry_out :
  selected_out == OUT_SUB ? ~adder_carry_out :
  selected_out == OUT_MUL ? 0 :
  0;

assign Fzero = ~((| O_lo) | (| O_hi));

// Modules

reg [31:0] adder_A, adder_B;
reg adder_carry_in, adder_carry_out;
wire [31:0] adder_out;

carry_select_adder adder(adder_A, adder_B, adder_carry_in, adder_out, adder_carry_out);

// Clocking

always @(posedge CLK) begin
  reg [63:0] mult_out_tmp;

  if (state == ST_IDLE && EN) begin
    case (op)
      OP_ADD: begin
        adder_A <= A;
        adder_B <= B;
        adder_carry_in <= 0;
        selected_out <= OUT_ADD;
      end
      OP_SUB: begin
        adder_A <= A;
        adder_B <= ~B;
        adder_carry_in <= 1;
        selected_out <= OUT_SUB;
      end
      OP_MUL: begin
        factorA <= A;
        factorB <= B;
        mult_out <= 0;
        adder_A <= 0;
        adder_B <= A;
        adder_carry_in <= 0;
        state <= ST_MULTIPLY_START;
        selected_out <= OUT_MUL;
      end
      OP_AND: begin
        bitwise_out <= A & B;
        selected_out <= OUT_BW;
      end
      OP_OR: begin
        bitwise_out <= A | B;
        selected_out <= OUT_BW;
      end
      OP_XOR: begin
        bitwise_out <= A ^ B;
        selected_out <= OUT_BW;
      end
      OP_NOT: begin
        bitwise_out <= ~A;
        selected_out <= OUT_BW;
      end
      default: begin end // TODO: this should be $stop in verilator, no-op in synthesis
    endcase
  end

  if (state >= ST_MULTIPLY_START && state <= ST_MULTIPLY_END) begin
    mult_out_tmp = {
      factorB[0] ? {adder_carry_out, adder_out} : {1'b0, mult_out[63:32]},
      mult_out[31:1]
    };
    mult_out <= mult_out_tmp;

    factorB <= {1'b0, factorB[31:1]};
    adder_A <= {mult_out_tmp[63:32]};
    adder_B <= factorA;
    if (state < ST_MULTIPLY_END) begin
      state <= state + 1;
    end else begin
      state <= ST_IDLE;
    end
  end
end

endmodule
